import json
from unittest.mock import MagicMock

import pytest

from pykube import Pod
from pykube.exceptions import HTTPError
from pykube.query import Query


@pytest.fixture
def api():
    return MagicMock()


def test_watch_response_exists(api):
    stream = Query(api, Pod).watch()
    assert hasattr(stream, "response")
    assert stream.response is None  # not yet executed


def test_watch_response_is_readonly(api):
    stream = Query(api, Pod).watch()
    with pytest.raises(AttributeError):
        stream.response = object()


def test_watch_response_is_set_on_iter(api):
    line1 = json.dumps({"type": "ADDED", "object": {}}).encode("utf-8")
    expected_response = MagicMock()
    expected_response.iter_lines.return_value = [line1]
    api.get.return_value = expected_response

    stream = Query(api, Pod).watch()
    next(iter(stream))

    assert stream.response is expected_response

    assert api.get.call_count == 1
    assert api.get.call_args_list[0][1]["stream"] is True
    assert "watch=true" in api.get.call_args_list[0][1]["url"]


def test_watch_exception_raised_on_failure(api):
    line1 = json.dumps(
        {
            "kind": "Status",
            "apiVersion": "meta.k8s.io/v1",
            "metadata": {},
            "status": "Failure",
            "message": "unexpected EOF",
            "code": 500,
        }
    ).encode("utf-8")
    expected_response = MagicMock()
    expected_response.iter_lines.return_value = [line1]
    api.get.return_value = expected_response

    stream = Query(api, Pod).watch()

    with pytest.raises(HTTPError):
        next(iter(stream))

    assert stream.response is expected_response

    assert api.get.call_count == 1
    assert api.get.call_args_list[0][1]["stream"] is True
    assert "watch=true" in api.get.call_args_list[0][1]["url"]


def test_watch_params_are_passed_through(api):
    line1 = json.dumps({"type": "ADDED", "object": {}}).encode("utf-8")
    expected_response = MagicMock()
    expected_response.iter_lines.return_value = [line1]
    api.get.return_value = expected_response

    params = dict(timeoutSeconds=123, arbitraryParam=456)
    stream = Query(api, Pod).watch(params=params)
    next(iter(stream))

    assert api.get.call_count == 1
    assert "timeoutSeconds=123" in api.get.call_args_list[0][1]["url"]
    assert "arbitraryParam=456" in api.get.call_args_list[0][1]["url"]



"""
pykube.http unittests
"""
import copy
import logging
import tempfile
from pathlib import Path

from . import TestCase

BASE_CONFIG = {
    "clusters": [
        {"name": "test-cluster", "cluster": {"server": "http://localhost:8080"}}
    ],
    "contexts": [
        {
            "name": "test-cluster",
            "context": {"cluster": "test-cluster", "user": "test-user"},
        }
    ],
    "users": [{"name": "test-user", "user": {}}],
    "current-context": "test-cluster",
}

_log = logging.getLogger(__name__)


class TestSession(TestCase):
    def setUp(self):
        self.config = copy.deepcopy(BASE_CONFIG)

    def test_build_session_auth_provider(self):
        """Test that HTTPClient correctly parses the auth-provider config.

        Observed in GKE with kubelet v1.3.
        """
        self.config.update(
            {
                "users": [
                    {
                        "name": "test-user",
                        "user": {
                            "auth-provider": {
                                "config": {
                                    "access-token": "abc",
                                    "expiry": "2016-08-24T16:19:17.19878675-07:00",
                                },
                            },
                        },
                    },
                ]
            }
        )

        gcloud_content = """
{
  "client_id": "myclientid",
  "client_secret": "myclientsecret",
  "refresh_token": "myrefreshtoken",
  "type": "authorized_user"
}

"""

        _log.info("Built config: %s", self.config)
        try:
            tmp = Path(tempfile.mktemp())

            with tmp.open("w") as f:
                f.write(gcloud_content)

            # TODO: this no longer works due to refactoring, GCP session handling is now done in KubernetesHTTPAdapter
            # session = pykube.session.GCPSession(pykube.KubeConfig(doc=self.config), tmp)
            # self.assertEquals(session.oauth.token['access_token'], 'abc')
            # self.assertEquals(session.oauth.token['refresh_token'], 'myrefreshtoken')
            # self.assertEquals(session.credentials.get('client_id'), 'myclientid')
            # self.assertEquals(session.credentials.get('client_secret'), 'myclientsecret')
        finally:
            if tmp.exists():
                tmp.unlink()



import io

import pytest

from pykube.console import main


@pytest.fixture
def kubeconfig(tmpdir):
    kubeconfig = tmpdir.join("kubeconfig")
    kubeconfig.write(
        """
apiVersion: v1
clusters:
- cluster: {server: 'https://localhost:9443'}
  name: test
contexts:
- context: {cluster: test}
  name: test-context
- context: {cluster: test}
  name: other
current-context: test-context
kind: Config
    """
    )
    return kubeconfig


def test_main_current_context(monkeypatch, capsys, kubeconfig):
    monkeypatch.setattr(
        "sys.stdin", io.StringIO('print(f"current context: {config.current_context}")')
    )
    main(["--kubeconfig", str(kubeconfig)])
    captured = capsys.readouterr()
    assert "current context: test-context" in captured.out


def test_main_custom_context(monkeypatch, capsys, kubeconfig):
    monkeypatch.setattr(
        "sys.stdin", io.StringIO('print(f"current context: {config.current_context}")')
    )
    main(["--kubeconfig", str(kubeconfig), "--context=other"])
    captured = capsys.readouterr()
    assert "current context: other" in captured.out


def test_main_script(monkeypatch, capsys, kubeconfig):
    monkeypatch.setattr("pykube.HTTPClient.version", ("1", "13"))
    main(["--kubeconfig", str(kubeconfig), "-c", "api.version"])
    captured = capsys.readouterr()
    assert "('1', '13')\n" == captured.out



"""
pykube.config unittests
"""
import os
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from . import TestCase
from pykube import config
from pykube import exceptions
from pykube.config import BytesOrFile

CERT_DUMMY = b"dummy"

BASEDIR = Path("tests")
GOOD_CONFIG_FILE_PATH = BASEDIR / "test_config.yaml"
DEFAULTUSER_CONFIG_FILE_PATH = BASEDIR / "test_config_default_user.yaml"


def test_from_service_account_no_file(tmpdir):
    with pytest.raises(FileNotFoundError):
        config.KubeConfig.from_service_account(path=str(tmpdir))


def test_from_service_account(tmpdir):
    namespace_file = Path(tmpdir) / "namespace"
    token_file = Path(tmpdir) / "token"
    ca_file = Path(tmpdir) / "ca.crt"

    with namespace_file.open("w") as fd:
        fd.write("mynamespace")

    with token_file.open("w") as fd:
        fd.write("mytok")

    with ca_file.open("w") as fd:
        fd.write("myca")

    os.environ["KUBERNETES_SERVICE_HOST"] = "127.0.0.1"
    os.environ["KUBERNETES_SERVICE_PORT"] = "9443"

    cfg = config.KubeConfig.from_service_account(path=str(tmpdir))

    assert cfg.doc["clusters"][0]["cluster"] == {
        "server": "https://127.0.0.1:9443",
        "certificate-authority": str(ca_file),
    }
    assert cfg.doc["users"][0]["user"]["token"] == "mytok"
    assert cfg.namespace == "mynamespace"


def test_from_url():
    cfg = config.KubeConfig.from_url("http://localhost:8080")
    assert cfg.doc["clusters"][0]["cluster"] == {"server": "http://localhost:8080"}
    assert "users" not in cfg.doc


@pytest.fixture
def kubeconfig(tmpdir):
    kubeconfig = tmpdir.join("kubeconfig")
    kubeconfig.write(
        """
apiVersion: v1
clusters:
- cluster: {server: 'https://localhost:9443'}
  name: test
contexts:
- context: {cluster: test, user: test}
  name: test
current-context: test
kind: Config
preferences: {}
users:
- name: test
  user: {token: testtoken}
    """
    )
    return kubeconfig


@pytest.mark.parametrize(
    "kubeconfig_env,expected_path",
    [(None, "~/.kube/config"), ("/some/path", "/some/path")],
)
def test_from_default_kubeconfig(
    kubeconfig_env, expected_path, monkeypatch, kubeconfig
):
    mock = MagicMock()
    mock.return_value.expanduser.return_value = Path(kubeconfig)
    monkeypatch.setattr(config, "Path", mock)

    if kubeconfig_env is None:
        monkeypatch.delenv("KUBECONFIG", raising=False)
    else:
        monkeypatch.setenv("KUBECONFIG", kubeconfig_env)

    cfg = config.KubeConfig.from_file()
    mock.assert_called_with(expected_path)
    assert cfg.doc["clusters"][0]["cluster"] == {"server": "https://localhost:9443"}


class TestConfig(TestCase):
    def setUp(self):
        self.cfg = config.KubeConfig.from_file(GOOD_CONFIG_FILE_PATH)

    def tearDown(self):
        self.cfg = None

    def test_init(self):
        """
        Test Config instance creation.
        """
        # Ensure that a valid creation works
        self.assertEqual(GOOD_CONFIG_FILE_PATH, self.cfg.filepath)

        # Ensure that if a file does not exist the creation fails
        self.assertRaises(
            exceptions.PyKubeError, config.KubeConfig.from_file, "doesnotexist"
        )

    def test_cert_temp_file_creation(self):
        """
        Verify the certificate data in the config is written to /tmp without leaks.
        https://codeberg.org/hjacobs/pykube-ng/issues/3
        """
        self.cfg.set_current_context("thename")

        num_cert_files = len(
            {
                self.cfg.cluster["certificate-authority"].filename(),
                self.cfg.cluster["certificate-authority"].filename(),
                self.cfg.cluster["certificate-authority"].filename(),
            }
        )

        self.assertEqual(1, num_cert_files)

    def test_set_current_context(self):
        """
        Verify set_current_context works as expected.
        """
        self.cfg.set_current_context("new_context")
        self.assertEqual("new_context", self.cfg.current_context)

    def test_clusters(self):
        """
        Verify clusters works as expected.
        """
        self.assertEqual(
            "http://localhost",
            self.cfg.clusters.get("thecluster", {}).get("server", None),
        )
        self.assertEqual(
            CERT_DUMMY,
            self.cfg.clusters.get("thecluster", {})
            .get("certificate-authority", {"_bytes": None})
            ._bytes,
        )

    def test_users(self):
        """
        Verify users works as expected.
        """
        self.__validate_user_certs(self.cfg.users.get("admin", {}))

    def test_contexts(self):
        """
        Verify contexts works as expected.
        """
        self.assertEqual(
            {"cluster": "thecluster", "user": "admin"},
            self.cfg.contexts.get("thename", None),
        )

    def test_cluster(self):
        """
        Verify cluster works as expected.
        """
        # Without a current_context this should fail
        try:
            cluster = self.cfg.cluster
            self.fail(
                "cluster was found without a current context set: {}".format(cluster)
            )
        except exceptions.PyKubeError:
            # We should get an error
            pass

        self.cfg.set_current_context("thename")
        self.assertEqual("http://localhost", self.cfg.cluster.get("server", None))
        self.assertEqual(
            CERT_DUMMY,
            self.cfg.cluster.get("certificate-authority", {"_bytes": None})._bytes,
        )

    def test_user(self):
        """
        Verify user works as expected.
        """
        # Without a current_context this should fail
        try:
            user = self.cfg.user
            self.fail("user was found without a current context set: {}".format(user))
        except exceptions.PyKubeError:
            # We should get an error
            pass

        self.cfg.set_current_context("thename")
        self.__validate_user_certs(self.cfg.user)

    def test_default_user(self):
        """
        User can sometimes be specified as 'default' with no corresponding definition
        """
        test_config = config.KubeConfig.from_file(DEFAULTUSER_CONFIG_FILE_PATH)
        test_config.set_current_context("a_context")
        self.assertIsNotNone(test_config.user)

    def test_namespace(self):
        self.cfg.set_current_context("thename")
        self.assertEqual("default", self.cfg.namespace)
        self.cfg.set_current_context("context_with_namespace")
        self.assertEqual("foospace", self.cfg.namespace)

    def __validate_user_certs(self, d: dict):
        key = d.get("client-key", None)
        data = d.get("client-certificate", None)

        self.assertIsNotNone("data", data)
        self.assertIsNotNone("data", key)

        self.assertIsInstance(data, BytesOrFile)
        self.assertIsInstance(key, BytesOrFile)

        self.assertEqual(CERT_DUMMY, data._bytes)
        self.assertEqual(CERT_DUMMY, key._bytes)



from pykube.utils import join_url_path
from pykube.utils import obj_merge


def test_obj_merge():
    assert obj_merge({}, {}) == {}
    assert obj_merge({"a": 1}, {}) == {"a": 1}
    assert obj_merge({}, {"b": 2}) == {"b": 2}
    assert obj_merge({"a": []}, {"a": []}) == {"a": []}
    assert obj_merge({"a": [1, 2]}, {"a": []}) == {"a": [1, 2]}
    assert obj_merge({"a": []}, {"a": [1, 2]}) == {"a": [1, 2]}
    assert obj_merge({"a": [1, 2]}, {"a": [3, 4]}) == {"a": [1, 2]}
    assert obj_merge({"a": {"b": [1, 2]}}, {"a": {"b": [3, 4, 5], "c": [1, 2]}}) == {
        "a": {"b": [1, 2, 5], "c": [1, 2]}
    }

    assert obj_merge(
        {"a": {"e": [1, 2], "f": [5, 6]}},
        {"a": {"e": [3, 4]}, "b": ["1"]},
        is_strategic=False,
    ) == {"a": {"e": [1, 2], "f": [5, 6]}}
    assert obj_merge({"a": []}, {"a": [1, 2]}, is_strategic=False) == {"a": []}
    assert obj_merge({"a": {"b": [1, 2]}}, {"a": [1, 2]}, is_strategic=False) == {
        "a": {"b": [1, 2]}
    }
    assert obj_merge(
        {"a": {"b": [1, 2]}}, {"a": {"b": [], "c": [1, 2]}}, is_strategic=False
    ) == {"a": {"b": [1, 2]}}
    assert obj_merge(
        {"a": {"b": [1, 2]}}, {"a": {"b": [3, 4, 5], "c": [1, 2]}}, is_strategic=False
    ) == {"a": {"b": [1, 2]}}


def test_join_url_path():
    assert join_url_path() == "/"
    assert join_url_path("") == "/"
    assert join_url_path("", "/") == "/"
    assert join_url_path("first", "") == "/first"
    assert join_url_path("first", "/") == "/first/"
    assert join_url_path("first", "second") == "/first/second"
    assert join_url_path("/first", "second/") == "/first/second"
    assert join_url_path("first", "second", "") == "/first/second"
    assert join_url_path("first", "/", "second", "", "") == "/first/second"
    assert join_url_path("/first", "second", "", join_empty=True) == "/first/second/"



"""
pykube.http unittests
"""
import copy
import logging

import pykube
from . import TestCase

BASE_CONFIG = {
    "clusters": [
        {"name": "test-cluster", "cluster": {"server": "http://localhost:8080"}}
    ],
    "contexts": [
        {
            "name": "test-cluster",
            "context": {"cluster": "test-cluster", "user": "test-user"},
        }
    ],
    "users": [{"name": "test-user", "user": {}}],
    "current-context": "test-cluster",
}

_log = logging.getLogger(__name__)


class TestHTTPClient(TestCase):
    def setUp(self):
        self.config = copy.deepcopy(BASE_CONFIG)

    def ensure_no_auth(self, client):
        self.assertIsNone(
            client.session.cert, msg="Should not send certs when not configured"
        )
        self.assertNotIn(
            "Authorization",
            client.session.headers,
            msg="Should not send basic auth when not configured",
        )

    def test_no_auth_with_empty_user(self):
        """
        Cluster does not require any authentication--so no credentials are provided in the user info
        """
        config = {
            "clusters": [
                {
                    "name": "no-auth-cluster",
                    "cluster": {"server": "http://localhost:8080"},
                }
            ],
            "users": [{"name": "no-auth-cluster", "user": {}}],
            "contexts": [
                {
                    "name": "no-auth-cluster",
                    "context": {
                        "cluster": "no-auth-cluster",
                        "user": "no-auth-cluster",
                    },
                }
            ],
            "current-context": "no-auth-cluster",
        }
        client = pykube.HTTPClient(pykube.KubeConfig(doc=config))
        self.ensure_no_auth(client)

    def test_no_auth_with_no_user(self):
        config = {
            "clusters": [
                {
                    "name": "no-auth-cluster",
                    "cluster": {"server": "http://localhost:8080"},
                }
            ],
            "contexts": [
                {"name": "no-auth-cluster", "context": {"cluster": "no-auth-cluster"}}
            ],
            "current-context": "no-auth-cluster",
        }
        client = pykube.HTTPClient(pykube.KubeConfig(doc=config))
        self.ensure_no_auth(client)

    def test_build_session_bearer_token(self):
        """Test that HTTPClient correctly parses the token"""
        self.config.update(
            {"users": [{"name": "test-user", "user": {"token": "test"}}]}
        )
        _log.info("Built config: %s", self.config)

        client = pykube.HTTPClient(pykube.KubeConfig(doc=self.config))
        _log.debug("Checking headers %s", client.session.headers)
        # TODO: session.headers is no long filled due to KubernetesHTTPAdapter!
        # self.assertIn('Authorization', client.session.headers)
        # self.assertEqual(client.session.headers['Authorization'], 'Bearer test')



import json
import operator

import pytest
import responses

import pykube
from pykube import Deployment
from pykube import HTTPClient
from pykube import KubeConfig
from pykube import ObjectDoesNotExist


@pytest.fixture
def kubeconfig(tmpdir):
    kubeconfig = tmpdir.join("kubeconfig")
    kubeconfig.write(
        """
apiVersion: v1
clusters:
- cluster: {server: 'https://localhost:9443'}
  name: test
contexts:
- context: {cluster: test, user: test}
  name: test
current-context: test
kind: Config
preferences: {}
users:
- name: test
  user: {token: testtoken}
    """
    )
    return kubeconfig


@pytest.fixture
def requests_mock():
    return responses.RequestsMock(target="pykube.http.KubernetesHTTPAdapter._do_send")


@pytest.fixture
def api(kubeconfig):
    config = KubeConfig.from_file(str(kubeconfig))
    return HTTPClient(config)


def test_api_version(api, requests_mock):
    with requests_mock as rsps:
        rsps.add(
            responses.GET,
            "https://localhost:9443/version/",
            json={"major": 1, "minor": 13},
        )
        assert api.version == (1, 13)


def test_get_ready_pods(api, requests_mock):
    # example from README
    with requests_mock as rsps:
        rsps.add(
            responses.GET,
            "https://localhost:9443/api/v1/namespaces/gondor-system/pods",
            json={
                "items": [
                    {"metadata": {"name": "pod-1"}, "status": {}},
                    {
                        "metadata": {"name": "pod-2"},
                        "status": {"conditions": [{"type": "Ready", "status": "True"}]},
                    },
                ]
            },
        )
        pods = pykube.Pod.objects(api).filter(namespace="gondor-system")
        ready_pods = list(filter(operator.attrgetter("ready"), pods))
        assert len(ready_pods) == 1
        assert ready_pods[0].name == "pod-2"


def test_get_pod_by_name(api, requests_mock):
    # example from README
    with requests_mock as rsps:
        rsps.add(
            responses.GET,
            "https://localhost:9443/api/v1/namespaces/gondor-system/pods/my-pod",
            json={"spec": {"containers": [{"image": "hjacobs/kube-janitor"}]}},
        )
        rsps.add(
            responses.GET,
            "https://localhost:9443/api/v1/namespaces/gondor-system/pods/other-pod",
            status=404,
        )

        pod = (
            pykube.Pod.objects(api).filter(namespace="gondor-system").get(name="my-pod")
        )
        assert pod.obj["spec"]["containers"][0]["image"] == "hjacobs/kube-janitor"

        pod = (
            pykube.Pod.objects(api)
            .filter(namespace="gondor-system")
            .get_or_none(name="my-pod")
        )
        assert pod.obj["spec"]["containers"][0]["image"] == "hjacobs/kube-janitor"

        pod = (
            pykube.Pod.objects(api)
            .filter(namespace="gondor-system")
            .get_by_name("my-pod")
        )
        assert pod.obj["spec"]["containers"][0]["image"] == "hjacobs/kube-janitor"

        pod = (
            pykube.Pod.objects(api)
            .filter(namespace="gondor-system")
            .get_or_none(name="other-pod")
        )
        assert pod is None


def test_selector_query(api, requests_mock):
    # example from README
    with requests_mock as rsps:
        rsps.add(
            responses.GET,
            "https://localhost:9443/api/v1/namespaces/gondor-system/pods?labelSelector=gondor.io%2Fname+in+%28api-web%2Capi-worker%29",
            json={"items": [{"meta": {}}]},
        )

        pods = pykube.Pod.objects(api).filter(
            namespace="gondor-system",
            selector={"gondor.io/name__in": {"api-web", "api-worker"}},
        )
        assert len(list(pods)) == 1

        rsps.add(
            responses.GET,
            "https://localhost:9443/api/v1/namespaces/default/pods?fieldSelector=status.phase%3DPending",
            json={"items": [{"meta": {}}]},
        )

        pending_pods = pykube.objects.Pod.objects(api).filter(
            field_selector={"status.phase": "Pending"}
        )

        assert len(list(pending_pods)) == 1


def test_create_delete_deployment(api, requests_mock):
    # example from README
    with requests_mock as rsps:
        rsps.add(
            responses.POST,
            "https://localhost:9443/apis/apps/v1/namespaces/gondor-system/deployments",
            json={},
        )

        obj = {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {"name": "my-deploy", "namespace": "gondor-system"},
            "spec": {
                "replicas": 3,
                "selector": {"matchLabels": {"app": "nginx"}},
                "template": {
                    "metadata": {"labels": {"app": "nginx"}},
                    "spec": {
                        "containers": [
                            {
                                "name": "nginx",
                                "image": "nginx",
                                "ports": [{"containerPort": 80}],
                            }
                        ]
                    },
                },
            },
        }
        pykube.Deployment(api, obj).create()

        rsps.add(
            responses.DELETE,
            "https://localhost:9443/apis/apps/v1/namespaces/gondor-system/deployments/my-deploy",
            json={},
        )

        obj = {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {"name": "my-deploy", "namespace": "gondor-system"},
        }
        pykube.Deployment(api, obj).delete()


def test_list_deployments(api, requests_mock):
    with requests_mock as rsps:
        rsps.add(
            responses.GET,
            "https://localhost:9443/apis/apps/v1/namespaces/default/deployments",
            json={"items": []},
        )

        assert list(Deployment.objects(api)) == []
        assert len(rsps.calls) == 1
        # ensure that we passed the token specified in kubeconfig..
        assert rsps.calls[0].request.headers["Authorization"] == "Bearer testtoken"


def test_list_and_update_deployments(api, requests_mock):
    with requests_mock as rsps:
        rsps.add(
            responses.GET,
            "https://localhost:9443/apis/apps/v1/namespaces/default/deployments",
            json={
                "items": [{"metadata": {"name": "deploy-1"}, "spec": {"replicas": 3}}]
            },
        )

        deployments = list(Deployment.objects(api))
        assert len(deployments) == 1
        deploy = deployments[0]
        assert deploy.name == "deploy-1"
        assert deploy.namespace == "default"
        assert deploy.replicas == 3

        deploy.replicas = 2

        rsps.add(
            responses.PATCH,
            "https://localhost:9443/apis/apps/v1/namespaces/default/deployments/deploy-1",
            json={"metadata": {"name": "deploy-1"}, "spec": {"replicas": 2}},
        )

        deploy.update()
        assert len(rsps.calls) == 2

        assert json.loads(rsps.calls[-1].request.body) == {
            "metadata": {"name": "deploy-1"},
            "spec": {"replicas": 2},
        }
        assert deploy.replicas == 2


def test_list_and_patch_deployments(api, requests_mock):
    with requests_mock as rsps:
        rsps.add(
            responses.GET,
            "https://localhost:9443/apis/apps/v1/namespaces/default/deployments",
            json={
                "items": [{"metadata": {"name": "deploy-1"}, "spec": {"replicas": 3}}]
            },
        )

        deployments = list(Deployment.objects(api))
        assert len(deployments) == 1
        deploy = deployments[0]
        assert deploy.name == "deploy-1"
        assert deploy.namespace == "default"
        assert deploy.replicas == 3

        deploy.replicas = 2

        rsps.add(
            responses.PATCH,
            "https://localhost:9443/apis/apps/v1/namespaces/default/deployments/deploy-1",
            json={
                "metadata": {"name": "deploy-1", "labels": {"a": "b"}},
                "spec": {"replicas": 2},
            },
        )

        deploy.patch({"metadata": {"labels": {"a": "b"}}})
        assert len(rsps.calls) == 2

        assert json.loads(rsps.calls[-1].request.body) == {
            "metadata": {"labels": {"a": "b"}}
        }
        assert deploy.labels["a"] == "b"


def test_pod_exists(api, requests_mock):
    with requests_mock as rsps:
        obj = {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {"name": "my-pod"},
            "spec": {"containers": []},
        }
        pod = pykube.Pod(api, obj)

        rsps.add(
            responses.GET,
            "https://localhost:9443/api/v1/namespaces/default/pods/my-pod",
            status=404,
        )
        assert not pod.exists()

        with pytest.raises(ObjectDoesNotExist):
            pod.exists(ensure=True)

        rsps.replace(
            responses.GET,
            "https://localhost:9443/api/v1/namespaces/default/pods/my-pod",
            json={},
        )
        assert pod.exists()


def test_reload(api, requests_mock):
    with requests_mock as rsps:
        rsps.add(
            responses.GET,
            "https://localhost:9443/api/v1/namespaces/default/pods/my-pod",
            json={"metadata": {"name": "my-pod", "labels": {"a": "foo"}}},
        )
        pod = pykube.Pod.objects(api).get_by_name("my-pod")
        assert pod.labels["a"] == "foo"

        rsps.replace(
            responses.GET,
            "https://localhost:9443/api/v1/namespaces/default/pods/my-pod",
            json={"metadata": {"name": "my-pod", "labels": {"a": "bar"}}},
        )
        pod.reload()
        assert pod.labels["a"] == "bar"


def test_resource_list(api, requests_mock):
    with requests_mock as rsps:
        data1 = {"resources": [{"kind": "Pod", "name": "pods"}]}
        rsps.add(responses.GET, "https://localhost:9443/api/v1/", json=data1)
        resource_list = api.resource_list("v1")
        assert resource_list == data1

        data2 = {"resources": [{"kind": "ExampleObject", "name": "exampleobjects"}]}
        rsps.add(
            responses.GET, "https://localhost:9443/apis/example.org/v1/", json=data2
        )
        resource_list = api.resource_list("example.org/v1")
        assert resource_list == data2


def test_patch_subresource(api, requests_mock):
    with requests_mock as rsps:
        rsps.add(
            responses.GET,
            "https://localhost:9443/apis/apps/v1/namespaces/default/deployments",
            json={"items": [{"metadata": {"name": "deploy-1"}}]},
        )

        deployments = list(Deployment.objects(api))
        assert len(deployments) == 1
        deploy = deployments[0]
        assert deploy.name == "deploy-1"
        assert deploy.namespace == "default"

        rsps.add(
            responses.PATCH,
            "https://localhost:9443/apis/apps/v1/namespaces/default/deployments/deploy-1/status",
            json={"metadata": {"name": "deploy-1"}, "status": {"field": "field"}},
        )

        deploy.patch({"status": {"field": "field"}}, subresource="status")
        assert len(rsps.calls) == 2

        assert json.loads(rsps.calls[-1].request.body) == {
            "status": {"field": "field"},
        }



from typing import Any
from typing import Dict
from typing import Optional
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockFixture

import pykube
from pykube.objects import NamespacedAPIObject
from pykube.objects import Pod
from pykube.utils import obj_merge


def test_api_object():
    pod = Pod(None, {"metadata": {"name": "myname"}})
    assert repr(pod) == "<Pod myname>"
    assert str(pod) == "myname"
    assert pod.metadata == {"name": "myname"}
    assert pod.labels == {}
    assert pod.annotations == {}


def test_object_factory_succeeds():
    api = MagicMock()
    api.resource_list.return_value = {
        "resources": [
            {"kind": "ExampleObject", "namespaced": True, "name": "exampleobjects"}
        ]
    }
    ExampleObject = pykube.object_factory(api, "example.org/v1", "ExampleObject")
    assert ExampleObject.kind == "ExampleObject"
    assert ExampleObject.endpoint == "exampleobjects"
    assert ExampleObject.version == "example.org/v1"
    assert NamespacedAPIObject in ExampleObject.mro()


def test_object_factory_raises_for_unknown_kind():
    api = MagicMock()
    api.resource_list.return_value = {
        "resources": [
            {"kind": "ExampleObject", "namespaced": True, "name": "exampleobjects"}
        ]
    }
    with pytest.raises(ValueError):
        pykube.object_factory(api, "example.org/v1", "OtherObject")


def test_set_annotation():
    pod = Pod(None, {"metadata": {"name": "myname"}})
    pod.annotations["foo"] = "bar"
    assert pod.annotations["foo"] == "bar"


def test_set_label():
    pod = Pod(None, {"metadata": {"name": "myname"}})
    pod.labels["foo"] = "bar"
    assert pod.labels["foo"] == "bar"


def test_update():
    pod = Pod(
        None,
        {"metadata": {"name": "john", "kind": "test"}, "annotations": "a long string"},
    )
    pod.obj = {"metadata": {"name": "john"}}
    pod.api = MagicMock()
    pod.api.patch.return_value.json.return_value = obj_merge(
        pod.obj, pod._original_obj, False
    )
    pod.update(is_strategic=False)
    assert pod.metadata == {"name": "john"}
    assert pod.annotations == {}


@pytest.mark.parametrize(
    "port,expected_request_kwargs",
    [
        # empty port - default service port should be used
        (
            None,
            {
                "url": "services/test_service:9090/proxy//",
                "namespace": "test_namespace",
                "version": "v1",
            },
        ),
        # override the default port, should be included in url
        (
            8080,
            {
                "url": "services/test_service:8080/proxy//",
                "namespace": "test_namespace",
                "version": "v1",
            },
        ),
    ],
)
def test_service_proxy_http_request(
    mocker: MockFixture, port: Optional[int], expected_request_kwargs: Dict[str, Any]
) -> None:
    mock_client = mocker.MagicMock(spec=pykube.http.HTTPClient)

    service = pykube.Service(
        mock_client,
        {
            "apiVersion": "core/v1",
            "kind": "Service",
            "metadata": {
                "name": "test_service",
                "namespace": "test_namespace",
            },
            "spec": {"ports": [{"port": 9090}]},
        },
    )
    service.proxy_http_request("GET", "/", port)

    request = mock_client.request
    assert request.called
    assert request.call_args_list[0][0] == ("GET",)
    assert request.call_args_list[0][1] == expected_request_kwargs



from unittest.mock import MagicMock

import pytest

from pykube import ObjectDoesNotExist
from pykube import Pod
from pykube.query import Query


@pytest.fixture
def api():
    return MagicMock()


def test_get(api):
    with pytest.raises(ObjectDoesNotExist):
        Query(api, Pod).get(namespace="myns")


def test_repr(api):
    query = Query(api, Pod)
    assert repr(query).startswith("<Query of Pod at ")


def test_get_one_object(api):
    response = MagicMock()
    response.json.return_value = {"items": [{"metadata": {"name": "pod1"}}]}
    api.get.return_value = response
    pod = Query(api, Pod).get(namespace="myns")
    assert pod.name == "pod1"


def test_get_more_than_one_object(api):
    response = MagicMock()
    response.json.return_value = {
        "items": [{"metadata": {"name": "pod1"}}, {"metadata": {"name": "pod2"}}]
    }
    api.get.return_value = response
    with pytest.raises(ValueError):
        Query(api, Pod).get(namespace="myns")


def test_filter_by_namespace(api):
    Query(api, Pod).filter(namespace="myns").execute()
    api.get.assert_called_once_with(namespace="myns", url="pods", version="v1")


def test_filter_by_labels_eq(api):
    Query(api, Pod).filter(
        selector={"application": "myapp", "component": "backend"}
    ).execute()
    api.get.assert_called_once_with(
        url="pods?labelSelector=application%3Dmyapp%2Ccomponent%3Dbackend", version="v1"
    )


def test_filter_by_labels_neq(api):
    Query(api, Pod).filter(selector={"application__neq": "myapp"}).execute()
    api.get.assert_called_once_with(
        url="pods?labelSelector=application+%21%3D+myapp", version="v1"
    )


def test_filter_by_labels_in(api):
    Query(api, Pod).filter(selector={"application__in": ["foo", "bar"]}).execute()
    api.get.assert_called_once_with(
        url="pods?labelSelector=application+in+%28bar%2Cfoo%29", version="v1"
    )


def test_filter_by_labels_notin(api):
    Query(api, Pod).filter(selector={"application__notin": ["foo", "bar"]}).execute()
    api.get.assert_called_once_with(
        url="pods?labelSelector=application+notin+%28bar%2Cfoo%29", version="v1"
    )


def test_filter_invalid_selector(api):
    with pytest.raises(ValueError):
        Query(api, Pod).filter(selector={"application__x": "foo"}).execute()


def test_filter_selector_string(api):
    Query(api, Pod).filter(selector="application=foo").execute()
    api.get.assert_called_once_with(
        url="pods?labelSelector=application%3Dfoo", version="v1"
    )


def test_as_table(api):
    response = MagicMock()
    response.json.return_value = {"kind": "Table", "columnDefinitions": [], "rows": []}
    api.get.return_value = response

    table = Query(api, Pod).filter(selector={"app": "foo"}).as_table()
    assert table.columns == []
    assert table.rows == []
    assert repr(table).startswith("<Table of Pod at")

    api.get.assert_called_once_with(
        url="pods?labelSelector=app%3Dfoo",
        version="v1",
        headers={"Accept": "application/json;as=Table;v=v1beta1;g=meta.k8s.io"},
    )



"""
pykube.http unittests
"""
import os
from pathlib import Path
from unittest import mock

import pytest

from pykube import __version__
from pykube.config import KubeConfig
from pykube.http import DEFAULT_HTTP_TIMEOUT
from pykube.http import HTTPClient


BASEDIR = Path("tests")
GOOD_CONFIG_FILE_PATH = BASEDIR / "test_config_with_context.yaml"
CONFIG_WITH_INSECURE_SKIP_TLS_VERIFY = (
    BASEDIR / "test_config_with_insecure_skip_tls_verify.yaml"
)
CONFIG_WITH_OVERRIDES = BASEDIR / "test_config_overrides.yaml"
CONFIG_WITH_OIDC_AUTH = BASEDIR / "test_config_with_oidc_auth.yaml"


def test_http(monkeypatch):
    cfg = KubeConfig.from_file(GOOD_CONFIG_FILE_PATH)
    api = HTTPClient(cfg)

    mock_send = mock.MagicMock()
    mock_send.side_effect = Exception("MOCK HTTP")
    monkeypatch.setattr("pykube.http.KubernetesHTTPAdapter._do_send", mock_send)

    with pytest.raises(Exception):
        api.get(url="test")

    mock_send.assert_called_once()
    assert (
        mock_send.call_args[0][0].headers["Authorization"]
        == "Basic YWRtOnNvbWVwYXNzd29yZA=="
    )
    assert mock_send.call_args[0][0].headers["User-Agent"] == f"pykube-ng/{__version__}"
    # check that the default HTTP timeout was set
    assert mock_send.call_args[1]["timeout"] == DEFAULT_HTTP_TIMEOUT


def test_http_with_dry_run(monkeypatch):
    cfg = KubeConfig.from_file(GOOD_CONFIG_FILE_PATH)
    api = HTTPClient(cfg, dry_run=True)

    mock_send = mock.MagicMock()
    mock_send.side_effect = Exception("MOCK HTTP")
    monkeypatch.setattr("pykube.http.KubernetesHTTPAdapter._do_send", mock_send)

    with pytest.raises(Exception):
        api.get(url="test")

    mock_send.assert_called_once()
    # check that dry run http parameters were set
    assert mock_send.call_args[0][0].url == "http://localhost/api/v1/test?dryRun=All"


def test_http_insecure_skip_tls_verify(monkeypatch):
    cfg = KubeConfig.from_file(CONFIG_WITH_INSECURE_SKIP_TLS_VERIFY)
    api = HTTPClient(cfg)

    mock_send = mock.MagicMock()
    mock_send.side_effect = Exception("MOCK HTTP")
    monkeypatch.setattr("pykube.http.KubernetesHTTPAdapter._do_send", mock_send)

    with pytest.raises(Exception):
        api.get(url="test")

    mock_send.assert_called_once()
    # check that SSL is not verified
    assert not mock_send.call_args[1]["verify"]


def test_http_override_certificate_authority(monkeypatch):
    cfg = KubeConfig.from_file(CONFIG_WITH_OVERRIDES)
    api = HTTPClient(cfg, dry_run=True)

    k = mock.patch.dict(
        os.environ, {"PYKUBE_SSL_CERTIFICATE_AUTHORITY": "/var/foo/bar/ca"}
    )
    k.start()
    mock_send = mock.MagicMock()
    mock_send.side_effect = Exception("MOCK HTTP")
    monkeypatch.setattr("pykube.http.KubernetesHTTPAdapter._do_send", mock_send)

    with pytest.raises(Exception):
        api.get(url="test")

    mock_send.assert_called_once()
    # Check path to overwritten CA is used
    assert mock_send.call_args[1]["verify"] == "/var/foo/bar/ca"
    k.stop()


def test_http_do_not_overwrite_auth(monkeypatch):
    cfg = KubeConfig.from_file(GOOD_CONFIG_FILE_PATH)
    api = HTTPClient(cfg)

    mock_send = mock.MagicMock()
    mock_send.side_effect = Exception("MOCK HTTP")
    monkeypatch.setattr("pykube.http.KubernetesHTTPAdapter._do_send", mock_send)

    with pytest.raises(Exception):
        api.get(url="test", headers={"Authorization": "Bearer testtoken"})

    mock_send.assert_called_once()
    assert mock_send.call_args[0][0].headers["Authorization"] == "Bearer testtoken"


def test_http_with_oidc_auth_no_refresh(monkeypatch):
    cfg = KubeConfig.from_file(CONFIG_WITH_OIDC_AUTH)
    api = HTTPClient(cfg)

    mock_send = mock.MagicMock()
    mock_send.side_effect = Exception("MOCK HTTP")
    monkeypatch.setattr("pykube.http.KubernetesHTTPAdapter._do_send", mock_send)

    with mock.patch(
        "pykube.http.KubernetesHTTPAdapter._is_valid_jwt", return_value=True
    ) as mock_jwt:
        with pytest.raises(Exception):
            api.get(url="test")
        mock_jwt.assert_called_once_with("some-id-token")

    mock_send.assert_called_once()
    assert mock_send.call_args[0][0].headers["Authorization"] == "Bearer some-id-token"


def test_get_kwargs():
    cfg = KubeConfig.from_file(GOOD_CONFIG_FILE_PATH)
    api = HTTPClient(cfg)

    assert api.get_kwargs(version="v1") == {
        "timeout": 10,
        "url": "http://localhost/api/v1/",
    }
    assert api.get_kwargs(version="/apis") == {
        "timeout": 10,
        "url": "http://localhost/apis/",
    }
    assert api.get_kwargs(version="storage.k8s.io/v1") == {
        "timeout": 10,
        "url": "http://localhost/apis/storage.k8s.io/v1/",
    }
